@@ -4,22 +4,25 @@
 
 """ Public XML parser module """
 
-import re
+import base64
+import hashlib
+import io
+import lxml.etree
 import os
-import codecs
+import re
+import requests
 import shutil
 import six
 import time
-import requests
-import lxml.etree
 import xml2rfc.log
 import xml2rfc.utils
+
 from xml2rfc.writers import base
 
 try:
-    from urllib.parse import urlparse, urljoin
+    from urllib.parse import urlparse, urljoin, urlsplit
 except ImportError:
-    from urlparse import urlparse, urljoin
+    from urlparse import urlparse, urljoin, urlsplit
 
 try:
     import debug
@@ -43,6 +46,9 @@ class XmlRfcError(Exception):
         self.filename = filename
         self.line = line_no
 
+    def __str__(self):
+        return self.msg
+
 class CachingResolver(lxml.etree.Resolver):
     """ Custom ENTITY request handler that uses a local cache """
     def __init__(self, cache_path=None, library_dirs=None, source=None,
@@ -116,7 +122,7 @@ class CachingResolver(lxml.etree.Resolver):
             path = os.path.join(dir, xml2rfc.CACHE_PREFIX)
             if os.access(path, os.W_OK):
                 shutil.rmtree(path)
-                xml2rfc.log.write('Deleted cache directory at', path)
+                xml2rfc.log.note('Deleted cache directory at', path)
 
                 
     def resolve(self, request, public_id, context):
@@ -133,19 +139,15 @@ class CachingResolver(lxml.etree.Resolver):
                 return self.resolve_string(self.rfc_number, context)
             return self.resolve_string("XXXX", context)
         if not urlparse(request).netloc:
-            # Format the request from the relative path of the source so that 
-            # We get the exact same path as in the XML document
             if request.startswith("file:///"):
-                request = request[8:]
-            try:
-                request = os.path.relpath(request, self.source_dir)
-            except ValueError as e:
-                xml2rfc.log.error(str(e))                
+                request = request[7:]
         try:
             path = self.getReferenceRequest(request)
+            return self.resolve_filename(path, context)
         except Exception as e:
             xml2rfc.log.error(str(e))
-        return self.resolve_filename(path, context)
+            return None
+
     
     def getReferenceRequest(self, request, include=False, line_no=0):
         """ Returns the correct and most efficient path for an external request
@@ -283,7 +285,6 @@ class CachingResolver(lxml.etree.Resolver):
                         #     result = os.path.join(self.source_dir, request)
                         #     attempts.append(result)
                 else:
-                    # Hanging filename
                     for dir in self.library_dirs:
                         # NOTE: Recursion can be implemented here
                         # Try local library directories
@@ -341,8 +342,10 @@ class CachingResolver(lxml.etree.Resolver):
 
             Checks for the existence of the cache and creates it if necessary.
         """
-        scheme, netloc, path, params, query, fragment = urlparse(url)
-        basename = os.path.basename(path)
+        scheme, netloc, path, query, fragment = urlsplit(url)
+        root, ext = os.path.splitext(path)
+        hash = '-'+base64.urlsafe_b64encode(hashlib.sha1(query.encode()).digest()).decode() if query else ''
+        basename = os.path.basename(root+hash+ext)
         typename = self.include and 'include' or 'entity'
         # Try to load the URL from each cache in `read_cache`
         for dir in self.read_caches:
@@ -354,25 +357,52 @@ class CachingResolver(lxml.etree.Resolver):
                 else:
                     xml2rfc.log.note('Resolving ' + typename + '...', url)
                     xml2rfc.log.note('Loaded from cache', cached_path)
-                    return cached_path
+                    xml = lxml.etree.parse(cached_path)
+                    if xml.getroot().tag != 'reference':
+                        return cached_path
+                    else:
+                        if self.validate_ref(xml):
+                            return cached_path
+                        else:
+                            xml2rfc.log.error('Failure validating reference xml from %s' % cached_path )
+                            os.path.unlink(cached_path)
+                            return url
 
         xml2rfc.log.note('Resolving ' + typename + '...', url)
         if not netloc in self.sessions:
             self.sessions[netloc] = requests.Session()
         session = self.sessions[netloc]
-        r = session.get(url)
+        exc = None
+        for i in range(4):
+            try:
+                r = session.get(url)
+                break
+            except requests.exceptions.ConnectionError as e:
+                exc = e
+                xml2rfc.log.note('  retrying %s (%s)' % (url, e.args[0].args[0]))
+        else:
+            xml2rfc.log.error('Failure fetching URL %s (%s)' % (url, exc.args[0].args[0]))
+        for rr in r.history + [r, ]:
+            xml2rfc.log.note(' ... %s %s' % (rr.status_code, rr.url))
         if r.status_code == 200:
             if self.write_cache:
                 try:
                     xml = lxml.etree.fromstring(r.text.encode('utf8'))
-                    xml.set('{%s}base'%xml2rfc.utils.namespaces['xml'], url)
-                    text = lxml.etree.tostring(xml, encoding='utf-8')
-                    write_path = os.path.join(self.write_cache, 
-                                              xml2rfc.CACHE_PREFIX, basename)
-                    with codecs.open(write_path, 'w', encoding='utf-8') as cache_file:
-                        cache_file.write(text.decode('utf-8'))
-                    xml2rfc.log.note('Added file to cache: ', write_path)
-                    return write_path
+                    if xml.tag != 'reference':
+                        return url
+                    else:
+                        if self.validate_ref(xml):
+                            xml.set('{%s}base'%xml2rfc.utils.namespaces['xml'], r.url)
+                            text = lxml.etree.tostring(xml, encoding='utf-8')
+                            write_path = os.path.join(self.write_cache, 
+                                                      xml2rfc.CACHE_PREFIX, basename)
+                            with io.open(write_path, 'w', encoding='utf-8') as cache_file:
+                                cache_file.write(text.decode('utf-8'))
+                            xml2rfc.log.note('Added file to cache: ', write_path)
+                            return write_path
+                        else:
+                            xml2rfc.log.error('Failure validating reference xml from %s' % url )
+                            return url
                 except Exception as e:
                     xml2rfc.log.error(str(e))
                     return url
@@ -383,9 +413,14 @@ class CachingResolver(lxml.etree.Resolver):
             xml2rfc.log.note("URL retrieval failed with status code %s for '%s'" % (r.status_code, r.url))
             return ''
 
+    def validate_ref(self, xml):
+        ref_rng_file = os.path.join(os.path.dirname(__file__), 'data', 'reference.rng')
+        ref_rng = lxml.etree.RelaxNG(file=ref_rng_file)
+        return ref_rng.validate(xml)
 
 class AnnotatedElement(lxml.etree.ElementBase):
     pis = None
+    page = None
     def get(self, key, default=None):
         value = super(AnnotatedElement, self).get(key, default)
         if value == default:
@@ -396,13 +431,13 @@ class AnnotatedElement(lxml.etree.ElementBase):
 class XmlRfcParser:
 
     nsmap = {
-        'xi':   'http://www.w3.org/2001/XInclude',
+        b'xi':   b'http://www.w3.org/2001/XInclude',
     }
 
 
     """ XML parser container with callbacks to construct an RFC tree """
     def __init__(self, source, verbose=None, quiet=None, options=base.default_options,
-                 cache_path=None, templates_path=None, library_dirs=None,
+                 cache_path=None, templates_path=None, library_dirs=None, add_xmlns=False,
                  no_network=None, network_locs=[
                      'https://xml2rfc.ietf.org/public/rfc/',
                      'https://xml2rfc.tools.ietf.org/public/rfc/',
@@ -419,12 +454,8 @@ class XmlRfcParser:
         self.network_locs = network_locs
 
         if self.source:
-            if six.PY2:
-                with open(self.source, "rU") as f:
-                    self.text = f.read()
-            else:
-                with open(self.source, "rb", newline=None) as f:
-                    self.text = f.read()
+            with io.open(self.source, "rb", newline=None) as f:
+                self.text = f.read()
 
         # Initialize templates directory
         self.templates_path = templates_path or \
@@ -443,12 +474,13 @@ class XmlRfcParser:
         #   2. Default to /usr/share/xml2rfc
         # Split on colon or semi-colon delimiters
         if not library_dirs:
-            library_dirs = os.environ.get('XML_LIBRARY', '/usr/share/xml2rfc:')
+            library_dirs = os.environ.get('XML_LIBRARY', '/usr/share/xml2rfc')
         self.library_dirs = []
-        for raw_dir in re.split(':|;', library_dirs):
+        srcdir = os.path.abspath(os.path.dirname(self.source))
+        for raw_dir in re.split(':|;', library_dirs) + [ srcdir ]:
             # Convert empty directory to source dir
             if raw_dir == '': 
-                raw_dir = os.path.abspath(os.path.dirname(self.source))
+                raw_dir = srcdir
             else:
                 raw_dir = os.path.normpath(os.path.expanduser(raw_dir))
             # Add dir if its unique
@@ -469,13 +501,18 @@ class XmlRfcParser:
     def delete_cache(self, path=None):
         self.cachingResolver.delete_cache(path=path)
 
-    def parse(self, remove_comments=True, remove_pis=False, quiet=False, strip_cdata=True, normalize=False):
+    def parse(self, remove_comments=True, remove_pis=False, quiet=False, strip_cdata=True, normalize=False, add_xmlns=False):
         """ Parses the source XML file and returns an XmlRfc instance """
-        if not (self.quiet or quiet):
-            xml2rfc.log.write('Parsing file', self.source)
+        xml2rfc.log.note('Parsing file', self.source)
+
+        # workaround for not being able to explicitly set namespaces on the
+        # xml root element in lxml: Insert it before we parse:
+        text = self.text
+        if add_xmlns and not self.nsmap[b'xi'] in text:
+            text = text.replace(b'<rfc ', b'<rfc xmlns:%s="%s" ' % (b'xi', self.nsmap[b'xi']), 1)
 
         # Get an iterating parser object
-        file = six.BytesIO(self.text)
+        file = six.BytesIO(text)
         context = lxml.etree.iterparse(file,
                                       dtd_validation=False,
                                       load_dtd=True,
@@ -554,7 +591,7 @@ class XmlRfcParser:
         parser.set_element_class_lookup(element_lookup)
 
         # Parse the XML file into a tree and create an rfc instance
-        file = six.BytesIO(self.text)
+        file = six.BytesIO(text)
         tree = lxml.etree.parse(file, parser)
         xmlrfc = XmlRfc(tree, self.default_dtd_path, nsmap=self.nsmap)
         xmlrfc.source = self.source
@@ -616,7 +653,7 @@ class XmlRfcParser:
 
         return xmlrfc
 
-class XmlRfc:
+class XmlRfc(object):
     """ Internal representation of an RFC document
 
         Contains an lxml.etree.ElementTree, with some additional helper
@@ -626,11 +663,14 @@ class XmlRfc:
     """
 
     pis = {}
+    source = "(unknown source)"
 
-    def __init__(self, tree, default_dtd_path, nsmap={}):
+    def __init__(self, tree, default_dtd_path, source=None, nsmap={}):
         self.default_dtd_path = default_dtd_path
         self.tree = tree
         self.nsmap = nsmap
+        if source:
+            self.source = source
         # Pi default values
         self.pis = {
             "artworkdelimiter":	None,
