@@ -2,11 +2,13 @@
 
 from __future__ import print_function
 
-import sys
+import io
 import os
+import six
+import sys
 
 # If this script is renamed to 'xml2rfc.py' on a Windows system, the import
-# of the realy xml2rfc module will break.  In order to handle this, we remove
+# of the real xml2rfc module will break.  In order to handle this, we remove
 # the directory of the script file from the python system path:
 script_dir = os.path.dirname(os.path.realpath(__file__))
 if script_dir in sys.path:
@@ -20,8 +22,8 @@ import os
 import xml2rfc
 
 try:
-    import debug
-    assert debug
+    from xml2rfc import debug
+    debug.debug = True
 except ImportError:
     pass
 
@@ -33,11 +35,6 @@ def display_version(self, opt, value, parser):
     sys.exit()
 
 
-def clear_cache(self, opt, value, parser):
-    xml2rfc.parser.XmlRfcParser('').delete_cache()
-    sys.exit()
-
-
 def print_pi_help(self, opt, value, parser):
     pis = xml2rfc.parser.XmlRfc(None, None).pis.items()
     pis.sort()
@@ -82,6 +79,12 @@ def main():
                            help='outputs to an html file')
     formatgroup.add_option('', '--nroff', action='store_true',
                            help='outputs to an nroff file')
+    if xml2rfc.HAVE_CAIRO and xml2rfc.HAVE_PANGO:
+        formatgroup.add_option('', '--pdf', action='store_true',
+                               help='outputs to a pdf file')
+    else:
+        formatgroup.add_option('', '--pdf', action='store_true',
+                               help='(unavailable due to missing external library)')
     formatgroup.add_option('', '--raw', action='store_true',
                            help='outputs to a text file, unpaginated')
     formatgroup.add_option('', '--expand', action='store_true',
@@ -96,26 +99,28 @@ def main():
 
 
     plain_options = optparse.OptionGroup(optionparser, 'Plain Options')
-    plain_options.add_option('-C', '--clear-cache', action='callback', callback=clear_cache,
+    plain_options.add_option('-C', '--clear-cache', action='store_true', default=False,
                             help='purge the cache and exit')
-    plain_options.add_option('-H', '--pi-help', action='callback', callback=print_pi_help,
-                            help='show the names and default values of PIs')
     plain_options.add_option(      '--debug', action='store_true',
                             help='Show debugging output')
+    plain_options.add_option('-H', '--pi-help', action='callback', callback=print_pi_help,
+                            help='show the names and default values of PIs')
     plain_options.add_option('-n', '--no-dtd', action='store_true',
                             help='disable DTD validation step')
     plain_options.add_option('-N', '--no-network', action='store_true', default=False,
                             help='don\'t use the network to resolve references')
+    plain_options.add_option('-O', '--no-org-info', dest='first_page_author_org', action='store_false', default=True,
+                            help='don\'t show author orgainzation info on page one (legacy only)')
+    plain_options.add_option('-r', '--remove-pis', action='store_true', default=False,
+                            help='Remove XML processing instructions')
     plain_options.add_option('-q', '--quiet', action='store_true',
-                            help='dont print anything')
+                            help="don't print anything")
     plain_options.add_option('-u', '--utf8', action='store_true',
                             help='generate utf8 output')
     plain_options.add_option('-v', '--verbose', action='store_true',
                             help='print extra information')
     plain_options.add_option('-V', '--version', action='store_true', 
                             help='display the version number and exit')
-    plain_options.add_option('-r', '--remove-pis', action='store_true', default=False,
-                            help='Remove XML processing instructions')
     optionparser.add_option_group(plain_options)
 
 
@@ -138,12 +143,13 @@ def main():
                             help='specify the directory path for output files')
     optionparser.add_option_group(value_options)
 
-
     formatoptions = optparse.OptionGroup(optionparser, 'Format Options')
     formatoptions.add_option('--v3', dest='legacy', action='store_false',
                            help='with --text and --html: use the v3 formatter, rather than the legacy one.')
     formatoptions.add_option('--legacy', default=True, action='store_true',
                            help='with --text and --html: use the legacy text formatter, rather than the v3 one.')
+    formatoptions.add_option('--id-is-work-in-progress', default=True, action='store_true',
+                           help='in references, refer to Internet-Drafts as "Work in Progress".')
     optionparser.add_option_group(formatoptions)
 
     textoptions = optparse.OptionGroup(optionparser, 'Text Format Options')
@@ -156,17 +162,27 @@ def main():
                            help='use the legacy date format, rather than the new one.')
     textoptions.add_option('--list-symbols', metavar='4*CHAR',
                            help='use the characters given as list bullet symbols.')
+    textoptions.add_option('--bom', '--BOM', action='store_true', default=False,
+                           help='Add a BOM (unicode byte order mark) to the start of text files')
+    textoptions.add_option('-P', '--no-pagination', dest='pagination', action='store_false', default=True,
+                            help='don\'t do pagination of v3 draft text format.  V3 RFC text output is never paginated.')
     optionparser.add_option_group(textoptions)
 
     htmloptions = optparse.OptionGroup(optionparser, 'Html Format Options')
     htmloptions.add_option('--css', default=None,
                            help='Use the given CSS file instead of the builtin')
     htmloptions.add_option('--external-css', action='store_true', default=False,
-                           help='place css in an external file')
-    htmloptions.add_option('--rfc-base-url', default="https://www.rfc-editor.org/info/",
+                           help='place css in external files')
+    htmloptions.add_option('--rfc-base-url', default="https://www.rfc-editor.org/rfc/",
                            help='Base URL for RFC links')
-    htmloptions.add_option('--id-base-url', default="https://www.ietf.org/archive/id/",
+    htmloptions.add_option('--id-base-url', default="https://tools.ietf.org/html/",
                            help='Base URL for Internet-Draft links')
+    htmloptions.add_option('--rfc-reference-base-url', default="https://rfc-editor.org/rfc/",
+                           help='Base URL for RFC reference targets, replacing the target="..." value given in the reference entry')
+    htmloptions.add_option('--id-reference-base-url', default="https://tools.ietf.org/html/",
+                           help='Base URL for I-D reference targets')
+    htmloptions.add_option('--metadata-js-url', default="metadata.min.js",
+                           help='URL for the metadata script')
     optionparser.add_option_group(htmloptions)
 
     v2v3options = optparse.OptionGroup(optionparser, 'V2-V3 Converter Options')
@@ -188,24 +204,113 @@ def main():
     (options, args) = optionparser.parse_args()
     # Some additional values not exposed as options
     options.doi_base_url = "https://doi.org/"
+    options.no_css = False
+    options.image_svg = False
+
+    # Check that the default_options have values for all options, for people
+    # calling xml2rfc library functions, rather than the command-line
+    from xml2rfc.writers.base import default_options
+    for key in options.__dict__:
+        if not key in default_options.__dict__:
+            sys.stderr.write("  Option missing from base.default_options: %s\n" % key)
+            sys.exit(2)
 
     # Show version information, then exit
     if options.version:
         print('%s %s' % (xml2rfc.NAME, xml2rfc.__version__))
         if options.verbose:
+            print('  Python %s' % sys.version.split()[0])
+            extras = set(['pycairo', 'weasyprint'])
             try:
                 import pkg_resources
                 this = pkg_resources.working_set.by_key[xml2rfc.NAME]
                 for p in this.requires():
-                    dist = pkg_resources.get_distribution(p.key)
-                    print('  %s'%dist)
+                    if p.key in extras:
+                        extras -= p.key
+                    try:
+                        dist = pkg_resources.get_distribution(p.key)
+                        print('  %s'%dist)
+                    except:
+                        pass
+                for key in extras:
+                    try:
+                        dist = pkg_resources.get_distribution(key)
+                        print('  %s'%dist)
+                    except:
+                        pass
             except:
                 pass
         sys.exit(0)
 
+    # Clear cache and exit if requested
+    if options.clear_cache:
+        xml2rfc.parser.XmlRfcParser('').delete_cache(path=options.cache)
+        sys.exit(0)
+
     if len(args) < 1:
         optionparser.print_help()
         sys.exit(2)
+
+    install_info = """
+    Cannot generate PDF due to missing external libraries.
+    ------------------------------------------------------
+    
+    In order to generate PDFs, xml2rfc uses the WeasyPrint library, which
+    depends on external libaries that must be installed as native packages.
+
+    1. First, install the Cairo, Pango, and GDK-PixBuf library files on your
+    system.  See installation instructions on the WeasyPrint Docs:
+    
+        https://weasyprint.readthedocs.io/en/stable/install.html
+
+    (Python 3 is not needed if your system Python is 2.7, though).
+
+
+    2. Next, install the pycairo and weasyprint python modules using pip.
+    Depending on your system, you may need to use 'sudo' or install in
+    user-specific directories, using the --user switch.  On OS X in
+    particular, you may also need to install a newer version of setuptools
+    using --user before weasyprint can be installed.  If you install with 
+    the --user switch, you may need to also set PYTHONPATH, e.g.,
+    
+        PYTHONPATH=/Users/username/Library/Python/2.7/lib/python/site-packages
+
+    for Python 2.7.
+
+    The basic pip commands (modify as needed according to the text above)
+    are:
+
+        pip install 'pycairo>=1.18' 'weasyprint<=0.42.3'
+
+
+    3. Finally, install the full Noto Font package:
+
+       * Download the full font file from:
+         https://noto-website-2.storage.googleapis.com/pkgs/Noto-unhinted.zip
+         or follow the 'DOWNLOAD ALL FONTS' link on this page:
+         https://www.google.com/get/noto/
+
+       * Follow the installation instructions at
+         https://www.google.com/get/noto/help/install/
+    
+    With these libraries, modules, and fonts installed and available to
+    xml2rfc, the --pdf switch will be enabled.
+    """
+
+    missing = ""
+    if options.pdf and not xml2rfc.HAVE_WEASYPRINT:
+        missing += "\nCould not import weasyprint"
+    if options.pdf and not xml2rfc.HAVE_PYCAIRO:
+        missing += "\nCould not import pycairo"
+    if options.pdf and not xml2rfc.HAVE_CAIRO:
+        missing += "\nCould not find the cairo lib"
+    if options.pdf and not xml2rfc.HAVE_PANGO:
+        missing += "\nCould not find the pango lib"
+
+    if missing:
+        install_info += missing + '\n'
+        sys.exit(install_info)
+
     source = args[0]
     if not os.path.exists(source):
         sys.exit('No such file: ' + source)
@@ -233,9 +338,9 @@ def main():
             options.output_path = options.basename
             options.basename = None
     #
-    num_formats = len([ o for o in [options.raw, options.text, options.nroff, options.html, options.expand, options.v2v3, options.preptool, options.info, ] if o])
+    num_formats = len([ o for o in [options.raw, options.text, options.nroff, options.html, options.expand, options.v2v3, options.preptool, options.info, options.pdf ] if o])
     if num_formats > 1 and (options.filename or options.output_filename):
-        sys.exit('Cannot give an explicit filename with more than one format, '
+        sys.exit('Cannot use an explicit output filename when generating more than one format, '
                  'use --path instead.')
     if num_formats < 1:
         # Default to paginated text output
@@ -247,9 +352,7 @@ def main():
         if not os.path.exists(options.cache):
             try:
                 os.makedirs(options.cache)
-                if options.verbose:
-                    xml2rfc.log.write('Created cache directory at', 
-                                      options.cache)
+                xml2rfc.log.note('Created cache directory at', options.cache)
             except OSError as e:
                 print('Unable to make cache directory: %s ' % options.cache)
                 print(e)
@@ -258,35 +361,20 @@ def main():
             if not os.access(options.cache, os.W_OK):
                 print('Cache directory is not writable: %s' % options.cache)
                 sys.exit(1)
+    #
     options.date = datetime.datetime.strptime(options.datestring, "%Y-%m-%d").date()
     if options.omit_headers and not options.text:
         sys.exit("You can only use --no-headers with paginated text output.")
     #
-    if options.text and not options.legacy:
-        if options.legacy_list_symbols and options.list_symbols:
-            sys.exit("You cannot specify both --list-symbols and --legacy_list_symbols.")
-        if options.list_symbols:
-            options.list_symbols = tuple(list(options.list_symbols))
-        elif options.legacy_list_symbols:
-            options.list_symbols = ('o', '*', '+', '-')
-        else:
-            options.list_symbols = ('*', '-', 'o', '+')
-    else:
-        if options.legacy_list_symbols:
-            sys.exit("You can only use --legacy_list_symbols with v3 text output.")
-        if options.list_symbols:
-            sys.exit("You can only use --list_symbols with v3 text output.")
-
-    if not options.legacy:
-        # I.e., V3 formatter
-        options.no_dtd = True        
-        if options.nroff:
-            sys.exit("You can only use --nroff in legacy mode")
-        if options.raw:
-            sys.exit("You can only use --raw in legacy mode")
-
     if options.utf8:
-        xml2rfc.log.warn("The --utf8 switch is deprecated.  Use the new unicode insertion element <u>\nto refer to unicode values in a protocol specification.")
+        xml2rfc.log.warn("The --utf8 switch is deprecated.  Use the new unicode insertion element <u> to refer to unicode values in a protocol specification.")
+
+    if options.rfc_reference_base_url:
+        if not options.rfc_reference_base_url.endswith('/'):
+            options.rfc_reference_base_url += '/'
+    if options.id_reference_base_url:
+        if not options.id_reference_base_url.endswith('/'):
+            options.id_reference_base_url += '/'
 
     # ------------------------------------------------------------------
 
@@ -310,18 +398,50 @@ def main():
         # matches lxml.etree._LogEntry, so we can use the same logging function
         xml2rfc.log.exception('Unable to parse the XML document: ' + args[0], e.error_log)
         sys.exit(1)
-        
+    # check doctype
+    if xmlrfc.tree.docinfo and xmlrfc.tree.docinfo.system_url:
+        version = xmlrfc.tree.getroot().get('version', '2')
+        if version == '3' and xmlrfc.tree.docinfo.system_url.lower() == 'rfc2629.dtd':
+            sys.exit('Incompatible schema information: found "rfc2629.dtd" in <DOCTYPE> of a version 3 file')
+
     # Remember if we're building an RFC
     options.rfc = xmlrfc.tree.getroot().get('number')
+    if options.rfc:
+        options.pagination = False
 
-    # Treat v3 input as v3
+    # Check if we've received a version="3" document, and adjust accordingly
     if xmlrfc.tree.getroot().get('version') == '3':
         options.legacy = False
-        options.vocabulary = 'v3'
         options.no_dtd = True
-        if options.list_symbols is None:
-            options.list_symbols = ('*', '-', 'o', '+')
+        options.vocabulary = 'v3'
 
+    # ------------------------------------------------------------------
+    # Additional option checks that depend on the option.legacy settin which
+    # we may have adjusted as a result of the <rfc version="..."> setting:
+    if options.text and not options.legacy:
+        if options.legacy_list_symbols and options.list_symbols:
+            sys.exit("You cannot specify both --list-symbols and --legacy_list_symbols.")
+        if options.list_symbols:
+            options.list_symbols = tuple(list(options.list_symbols))
+        elif options.legacy_list_symbols:
+            options.list_symbols = ('o', '*', '+', '-')
+        else:
+            options.list_symbols = ('*', '-', 'o', '+')
+    else:
+        if options.legacy_list_symbols:
+            sys.exit("You can only use --legacy-list-symbols with v3 text output.")
+        if options.list_symbols:
+            sys.exit("You can only use --list-symbols with v3 text output.")
+
+    if not options.legacy:
+        # I.e., V3 formatter
+        options.no_dtd = True        
+        if options.nroff:
+            sys.exit("You can only use --nroff in legacy mode")
+        if options.raw:
+            sys.exit("You can only use --raw in legacy mode")
+
+    # ------------------------------------------------------------------
     # Validate the document unless disabled
     if not options.no_dtd:
         ok, errors = xmlrfc.validate(dtd_path=options.dtd)
@@ -417,7 +537,7 @@ def main():
         # --- End of legacy formatter invocations ---
 
         if options.expand and not options.legacy:
-            xmlrfc = parser.parse(remove_comments=False, quiet=True, normalize=False, strip_cdata=False)
+            xmlrfc = parser.parse(remove_comments=False, quiet=True, normalize=False, strip_cdata=False, add_xmlns=True)
             filename = options.output_filename
             if not filename:
                 filename = basename + '.exp.xml'
@@ -429,7 +549,7 @@ def main():
             options.output_filename = None
 
         if options.v2v3:
-            xmlrfc = parser.parse(remove_comments=False, quiet=True, normalize=False)
+            xmlrfc = parser.parse(remove_comments=False, quiet=True, normalize=False, add_xmlns=True)
             filename = options.output_filename
             if not filename:
                 filename = basename + '.v2v3.xml'
@@ -439,7 +559,7 @@ def main():
             options.output_filename = None
 
         if options.preptool:
-            xmlrfc = parser.parse(remove_comments=False, quiet=True)
+            xmlrfc = parser.parse(remove_comments=False, quiet=True, add_xmlns=True)
             filename = options.output_filename
             if not filename:
                 filename = basename + '.prepped.xml'
@@ -451,7 +571,7 @@ def main():
             options.output_filename = None
 
         if options.text and not options.legacy:
-            xmlrfc = parser.parse(remove_comments=False, quiet=True)
+            xmlrfc = parser.parse(remove_comments=False, quiet=True, add_xmlns=True)
             filename = options.output_filename
             if not filename:
                 filename = basename + '.txt'
@@ -460,12 +580,13 @@ def main():
             xmlrfc.tree = v2v3.convert2to3()
             prep = xml2rfc.PrepToolWriter(xmlrfc, options=options, date=options.date, liberal=True, keep_pis=[xml2rfc.V3_PI_TARGET])
             xmlrfc.tree = prep.prep()
-            writer = xml2rfc.TextWriter(xmlrfc, options=options, date=options.date)
-            writer.write(filename)
-            options.output_filename = None
+            if xmlrfc.tree:
+                writer = xml2rfc.TextWriter(xmlrfc, options=options, date=options.date)
+                writer.write(filename)
+                options.output_filename = None
 
         if options.html and not options.legacy:
-            xmlrfc = parser.parse(remove_comments=False, quiet=True)
+            xmlrfc = parser.parse(remove_comments=False, quiet=True, add_xmlns=True)
             filename = options.output_filename
             if not filename:
                 filename = basename + '.html'
@@ -474,9 +595,25 @@ def main():
             xmlrfc.tree = v2v3.convert2to3()
             prep = xml2rfc.PrepToolWriter(xmlrfc, options=options, date=options.date, liberal=True, keep_pis=[xml2rfc.V3_PI_TARGET])
             xmlrfc.tree = prep.prep()
-            writer = xml2rfc.HtmlWriter(xmlrfc, options=options, date=options.date)
-            writer.write(filename)
-            options.output_filename = None
+            if xmlrfc.tree:
+                writer = xml2rfc.HtmlWriter(xmlrfc, options=options, date=options.date)
+                writer.write(filename)
+                options.output_filename = None
+
+        if options.pdf:
+            xmlrfc = parser.parse(remove_comments=False, quiet=True, add_xmlns=True)
+            filename = options.output_filename
+            if not filename:
+                filename = basename + '.pdf'
+                options.output_filename = filename
+            v2v3 = xml2rfc.V2v3XmlWriter(xmlrfc, options=options, date=options.date)
+            xmlrfc.tree = v2v3.convert2to3()
+            prep = xml2rfc.PrepToolWriter(xmlrfc, options=options, date=options.date, liberal=True, keep_pis=[xml2rfc.V3_PI_TARGET])
+            xmlrfc.tree = prep.prep()
+            if xmlrfc.tree:
+                writer = xml2rfc.PdfWriter(xmlrfc, options=options, date=options.date)
+                writer.write(filename)
+                options.output_filename = None
 
         if options.info:
             xmlrfc = parser.parse(remove_comments=False, quiet=True)
@@ -488,15 +625,21 @@ def main():
             xmlrfc.tree = v2v3.convert2to3()
             prep = xml2rfc.PrepToolWriter(xmlrfc, options=options, date=options.date, liberal=True, keep_pis=[xml2rfc.V3_PI_TARGET])
             xmlrfc.tree = prep.prep()
-            info = extract_anchor_info(xmlrfc.tree)
-            with open(filename, 'w') as fp:
-                json.dump(info, fp, indent=2, ensure_ascii=False, encoding='utf-8')
-            if not options.quiet:
-                xml2rfc.log.write('Created file', filename)
+            if xmlrfc.tree:
+                info = extract_anchor_info(xmlrfc.tree)
+                if six.PY2:
+                    with open(filename, 'w') as fp:
+                        json.dump(info, fp, indent=2, ensure_ascii=False, encoding='utf-8')
+                else:
+                    with io.open(filename, 'w', encoding='utf-8') as fp:
+                        json.dump(info, fp, indent=2, ensure_ascii=False)
+                if not options.quiet:
+                    xml2rfc.log.write('Created file', filename)
 
     except xml2rfc.RfcWriterError as e:
-        xml2rfc.log.error('Unable to convert the document: ' + args[0],  
-                          '\n  ' + e.msg)
+        xml2rfc.log.write(e.msg)
+        xml2rfc.log.write('Unable to complete processing %s' % args[0])
+        sys.exit(1)
 
 if __name__ == '__main__':
 
