@@ -9,7 +9,7 @@ import re
 import xml2rfc.log
 
 try:
-    import debug
+    from xml2rfc import debug
     debug.debug = True
 except ImportError:
     pass
@@ -53,6 +53,12 @@ def get_value(e, latin=True):
         value = e.text or ''
     return value
 
+country_alias = {
+    "UK": "United Kingdom",
+    "Republic of Korea": "Korea, Republic of",
+    "South Korea": "Korea, Republic of",
+}
+
 def get_iso_country_info(e):
     country_info = None
     ascii = e.get('ascii')
@@ -63,8 +69,8 @@ def get_iso_country_info(e):
     else:
         return None
     # Special case, this is used bu unknown to pycountry
-    if name == 'UK':
-        name = 'United Kingdom'
+    if name in country_alias:
+        name = country_alias[name]
     try:
         if hasattr(pycountry.countries, 'lookup'):
             country_info = pycountry.countries.lookup(name)
@@ -100,7 +106,9 @@ def get_normalized_address_info(writer, x, latin=True):
                 break
     if country_info:
         country_name = get_value(country_element, latin=latin)
-        if re.match('[A-Z]{2,3}', country_name):
+        if country_name in ['US', 'USA']:
+            country_name = country_info.official_name
+        elif re.match('[A-Z]{2,3}', country_name):
             country_name = country_info.name
         adr = {
                 'name': name,
@@ -117,7 +125,6 @@ def get_normalized_address_info(writer, x, latin=True):
                 'country_name': country_name,
             }
     else:
-        writer.warn(x, "Did not find a recognized country entry for author %s" % full_author_name(author, latin))
         adr = {
                 'name': name,
                 'role': role,
@@ -172,7 +179,7 @@ def get_normalized_address_info(writer, x, latin=True):
             writer.note(x, "Postal address check failed for author %s." % full_author_name(author, latin))
             for item in e.errors:
                 if adr[item]:
-                    writer.note(x, "  Postal address has unexpected address info: %s: %s" % (address_field_elements[item], (adr[item])), label='')
+                    writer.note(x, "  Postal address has unexpected address field or field value: %s: %s" % (address_field_elements[item], (adr[item])), label='')
                 else:
                     writer.note(x, "  Postal address is missing an address element: %s" % (address_field_elements[item], ), label='')
         if list_parts:
@@ -206,17 +213,30 @@ def get_address_format_rules(address, latin=False):
     return address_format, rules
 
 def format_address(address, latin=False):
+    def hasword(line):
+        return re.search(r'\w', line, re.U) != None
     address_format, rules = get_address_format_rules(address, latin)
     address_line_formats = address_format.split('%n')
     address_lines = [
         _format_address_line(lf, address, rules)
         for lf in address_line_formats]
-    address_lines = filter(None, address_lines)
+    address_lines = filter(hasword, address_lines)
     return '\n'.join(address_lines)
 
 def enhance_address_format(rules, address_format):
-    # Add extended address field (fails for 4 countries: Guinea, Hungary,
-    # Iran and China):
+    #     'A': 'street_address'
+    #     'C': 'city'
+    #     'D': 'city_area'
+    #     'E': 'extended_address'
+    #     'N': 'name'
+    #     'O': 'company_name'
+    #     'S': 'country_area'
+    #     'X': 'sorting_code'
+    #     'Y': 'country_name '
+    #     'Z': 'postal_code'
+
+    # Add extended address field '%E' (the code fails for 4 countries: Guinea,
+    # Hungary, Iran and China, which won't have %E support):
     address_format = address_format.replace('%N%n%O%n%A', '%N%n%O%n%E%n%A')
     address_format = address_format.replace('%A%n%O%n%N', '%A%n%E%n%O%n%N')
     address_format = address_format.replace('%O%n%N%n%A', '%O%n%N%n%E%n%A')
@@ -224,6 +244,14 @@ def enhance_address_format(rules, address_format):
         address_format = '%Y%n' + address_format
     else:
         address_format = address_format + '%n%Y'
+    # This is incorrect, but suits us better when listing authors
+    # in the Authors' Addresses section: move the name to the first line
+    address_format = '%N%n' + re.sub(r'%N(%n)?', '', address_format)
+    # Get rid of any trailing newlines we might have created
+    address_format = re.sub(r'%n$', '', address_format)
+    # Add some spacing, for a few special cases
+    address_format = re.sub(r'%([A-Z])%([A-Z])', r'%\1 %\2', address_format)
+    address_format = re.sub(r'%([A-Z])%([A-Z])', r'%\1 %\2', address_format)
     # country-specific fixes, if any
     #if rules['country_code'] == 'SE':
     #    pass
