@@ -10,8 +10,8 @@ import six
 import unicodedata
 
 try:
-    import debug
-    debug.debug = True
+    from xml2rfc import debug
+    assert debug
 except ImportError:
     pass
 
@@ -46,6 +46,8 @@ for t in bare_unicode_tags:
     assert t in unicode_content_tags
 
 def expand_unicode_element(e, bare=False):
+    if not e.text:
+        return ''
     ascii = e.get('ascii')
     format = e.get('format', 'lit-name-num')
     size = len(e.text)
@@ -66,7 +68,7 @@ def expand_unicode_element(e, bare=False):
             for c in e.text:
                 if isinstance(c, six.binary_type):
                     c = six.text_type(c, encoding='latin1')
-                names.append(unicodedata.name(c))
+                names.append(unicodedata.name(c, 'U+%04x'%ord(c)))
             return ', '.join(names)
         except ValueError as exc:
             raise ValueError('When looking for the name of "%s", got: %s' % (c, exc))
@@ -91,8 +93,11 @@ def expand_unicode_element(e, bare=False):
         parts = re.findall('{([a-z]+)}', format)
     if len(parts) == 0 or not all(p for p in permitted):
         raise ValueError('Expected <unicode> expansion parts to be one or more of %s, but found format="%s"' % (', '.join(permitted), val))
-    if not any(p for p in required):
-        raise ValueError('Expected <unicode> expansion parts to include at least one of %s, but found format="%s"' % (', '.join(required), val))
+    if not any(p in parts for p in required):
+        if len(required) > 1:
+            raise ValueError('Expected <unicode> expansion parts to include at least one of %s, but found format="%s"' % (', '.join(['"%s"'%r for r in required]), val))
+        else:
+            raise ValueError('Expected <unicode> expansion parts to include at least "%s", but found format="%s"' % (required[0], val))            
     if len(parts) > 3:
         raise ValueError('Expected up to 3 dash-separated <unicode> expansion parts, but found %d: format="%s"' % (len(parts), format))
     #
@@ -126,7 +131,7 @@ def isascii(u):
         return True
     if isinstance(u, six.text_type):
         t = u+''
-        for ch in [ '\u00a0', '\u2011', '\u2028', ]:
+        for ch in [ '\u00a0', '\u200B', '\u2011', '\u2028', '\u2060', ]:
             if ch in t:
                 t = t.replace(ch, ' ')
         try:
@@ -137,6 +142,10 @@ def isascii(u):
     else:
         return True
 
+def textwidth(u):
+    "Length of string, disregarding zero-width code points"
+    return len(re.sub('[\u200B\u200C\u2060\uE060]', '', u))
+
 def downcode_punctuation(str):
     return downcode(str, replacements=punctuation)
 
@@ -159,7 +168,7 @@ def downcode(str, replacements=None):
         replacements = unicode_replacements
 
     while True:
-        match = re.search(u'([^ -\x7e\u2060\u200B\u00A0\u2011\u2028\r\n])', str)
+        match = re.search(u'([^ -\x7e\u2060\u200B\u00A0\u2011\u2028\uE060\r\n])', str)
         if not match:
             return str
         if match.group(1) in unicode_replacements:
