@@ -7,8 +7,10 @@ from __future__ import unicode_literals, print_function, division
 import base64
 import re
 import six
+import sys
 import textwrap
 
+from collections import OrderedDict
 from lxml.etree import _Comment
 
 if six.PY2:
@@ -39,8 +41,9 @@ class TextWrapper(textwrap.TextWrapper):
             u'[ \t\n\r\f\v]+|'                                  # any ASCII whitespace
             u'[^\\s-]*\\w+/(?=[A-Za-z]\\w*)|'                   # forward-slash separated words
             u'[^\\s-]*\\w+[^0-9\\s]-(?=\\w+[^0-9\\s])|'         # hyphenated words
+            u'\u200b|'                                          # &zwsp; zero-width space is breakable space
             u'''(?<=[\\w\\!"'\\&\\.\\,\\?])-{2,}(?=\\w))'''     # em-dash
-            u'(?![\u2060|\u200B])')                             # UNLESS &wj; or &zwbs; or &nbsp;
+            u'(?![\u00A0|\u2060|\uE060])')                      # UNLESS &nbsp; or &wj; 
 
         self.wordsep_re_uni = re.compile(self.wordsep_re.pattern, re.U)
 
@@ -68,15 +71,6 @@ class TextWrapper(textwrap.TextWrapper):
         # Start of next sentence regex
         self.sentence_start_re = re.compile("^[\"'([]*[A-Z]")
 
-        # XmlCharRef replacements that occur AFTER line breaking logic
-        self.post_break_replacements = {
-            u'\u2060': '',    # wj
-            u'\u200B': '',    # zwsp
-        }
-
-        self.break_on_hyphens = True
-
-
     def _fix_sentence_endings(self, chunks):
         """_fix_sentence_endings(chunks : [string])
 
@@ -99,19 +93,23 @@ class TextWrapper(textwrap.TextWrapper):
             else:
                 i += 1
 
-    def replace(self, text):
-        """ Replace control entities with the proper character 
-            after breaking has occured.
-        """
-        for key, val in self.post_break_replacements.items():
-            text = re.sub(re.escape(key), val, text)
-        return text
-
     def wrap(self, text, initial='', subsequent_indent=None, width=None,
         fix_doublespace=True, fix_sentence_endings=True, drop_whitespace=True):
         """ Mirrored implementation of wrap which replaces characters properly
             and also lets you easily specify indentation on the fly
         """
+        # Handle U+2028 LINE SEPARATOR by recursing
+        if '\u2028' in text:
+            parts = text.split('\u2028')
+            wrapped = []
+            for part in parts:
+                part = part.lstrip()
+                wrapped += self.wrap(part, initial=initial, subsequent_indent=subsequent_indent,
+                                    width=width, fix_doublespace=fix_doublespace,
+                                    fix_sentence_endings=fix_sentence_endings,
+                                    drop_whitespace=drop_whitespace)
+            return wrapped
+
         # Set indentation
         _width = self.width
         if width != None:
@@ -124,8 +122,9 @@ class TextWrapper(textwrap.TextWrapper):
         self.fix_sentence_endings = fix_sentence_endings
         self.drop_whitespace = drop_whitespace
 
-        # Original implementation
+        # --- Original implementation ----------------------------------
         text = self._munge_whitespace(text)
+        # --- End original implementation ------------------------------
 
         # Maybe remove double (and more) spaces, except when they might be between sentences
         if fix_doublespace:
@@ -136,16 +135,24 @@ class TextWrapper(textwrap.TextWrapper):
         text = re.sub("(Section|Appendix|Figure|Table) ", u"\\1\u00A0", text)
 
         # Replace some characters after splitting has occured
-        parts = self._split(text)
+        # --- Original implementation ----------------------------------
+        chunks = self._split(text)
+        # --- End original implementation ------------------------------
+
+        parts = chunks
         chunks = []
         max_word_len = self.width - len(self.subsequent_indent)
         for chunk in parts:
-            chunk2 = self.replace(chunk)
+            # remove zero-width characters before looking at length
+            chunk2 = re.sub('[\u200B\u200C\u2060\uE060]', '', chunk)
             if len(chunk2) > max_word_len:
-                chunk2 = chunk.replace(u'\u200B', '')
+                # No use trying not to break when we have a chunk longer
+                # than available horizontal space.  Remove wj and nbsp and
+                # try again.
+                chunk2 = re.sub('[\u2060\uE060]', '', chunk).replace('\u00A0', ' ')
                 bits = self._split(chunk2)
                 for bit in bits:
-                    chunk3 = self.replace(bit)
+                    chunk3 = re.sub('[\u2060\uE060]', '', bit).replace('\u00A0', ' ')
                     if len(chunk3) > max_word_len:
                         chunks += self._split(chunk3)
                     else:
@@ -153,15 +160,38 @@ class TextWrapper(textwrap.TextWrapper):
             else:
                 chunks += [ chunk2 ]
         
-        # Original implementation
+        # --- Original implementation ----------------------------------
         if self.fix_sentence_endings:
             self._fix_sentence_endings(chunks)
         wrapped = self._wrap_chunks(chunks)
+        # --- End original implementation ------------------------------
+
         self.width = _width
+        for i, chunk in enumerate(wrapped):
+            wrapped[i] = chunk.replace(u'\uE060', '')
         return wrapped
 
     def fill(self, *args, **kwargs):
-        return "\n".join(self.wrap(*args, **kwargs)).replace(u'\u00A0', ' ')
+        kwargs.pop('fill', None)
+        return "\n".join(self.wrap(*args, **kwargs))
+
+
+class TextSplitter(textwrap.TextWrapper):
+    """ Subclass that overrides a few things in the standard implementation """
+    def __init__(self, **kwargs):
+        textwrap.TextWrapper.__init__(self, **kwargs)
+
+        # Override wrapping regex, preserve '/' before linebreak
+        self.wordsep_re = re.compile(
+            u'('
+            u'[ \t\n\r\f\v]+|'                                  # any ASCII whitespace
+            u'[^\\s-]*\\w+/(?=[A-Za-z]\\w*)|'                   # forward-slash separated words
+            u'\u200b|'                                          # &zwsp; zero-width space is breakable space
+            u'\u2028|'                                          # &br; unicode 'line separator'
+            u'''(?<=[\\w\\!"'\\&\\.\\,\\?])-{2,}(?=\\w))'''     # em-dash
+            u'(?![\u00A0|\u2060|\uE060])')                      # UNLESS &nbsp; or &wj; 
+
+        self.wordsep_re_uni = re.compile(self.wordsep_re.pattern, re.U)
 
 
 def justify_inline(left_str, center_str, right_str, width=72):
@@ -207,22 +237,24 @@ def ascii_split(text):
     return re.split("[ \t\n\r\f\v]+", text)
 
 
-def urlkeep(text):
+def urlkeep(text, max=72):
     """ Insert word join XML entities on forward slashes and hyphens
         in a URL so that it stays on one line
     """
-    wj_char = u'\u2060'
-    zwsp_char = u'\u200B'
+    wj_char = u'\uE060'
     def replacer(match):
-        return match.group(0).replace('/', '/' + zwsp_char) \
-                             .replace('/' + zwsp_char + '/' + zwsp_char, '/' + wj_char +'/' + wj_char) \
-                             .replace('-', '-' + wj_char) \
-                             .replace(':', ':' + wj_char)
-                             
+        url = match.group(0)
+        if len(url) > max:
+            return url
+        else:
+            return ( url.replace('/', '/' + wj_char) 
+                        .replace('-', '-' + wj_char) 
+                        .replace(':', ':' + wj_char)
+                    )
     if 'http://' in text:
-        return re.sub('(?<=http:)\S*', replacer, text)
+        text = re.sub('http:\S*', replacer, text)
     if 'https://' in text:
-        return re.sub('(?<=https:)\S*', replacer, text)
+        text = re.sub('https:\S*', replacer, text)
     return text
 
 # ----------------------------------------------------------------------
@@ -280,9 +312,11 @@ def safeTagSlashedWords(tree):
     """
     slashList = {}
     for element in _slash_replacements:
-        slashList[element] = re.sub(u'/', u'\u200B/\u200B', element)
+        slashList[element] = re.sub(u'/', u'\uE060/\uE060', element)
         
     for element in tree.iter():
+        if element.tag == 'artwork':
+            continue
         if element.text:
             element.text = _replace_slashed_words(element.text, slashList)
         if element.tail:
@@ -324,6 +358,7 @@ def _replace_unicode_characters(str):
     """ replace those Unicode characters that we do not use internally
         &wj; &zwsp; &nbsp; &nbhy;
     """
+    str = str.replace('\uE060', '\u2060')
     while True:
         match = re.search(u'([^ -\x7e\u2060\u200B\u00A0\u2011\r\n])', str)
         if not match:
@@ -558,11 +593,25 @@ def build_dataurl(mime, data, base64enc=False):
 namespaces={
     'x':    'http://relaxng.org/ns/structure/1.0',
     'a':    'http://relaxng.org/ns/compatibility/annotations/1.0',
+    'xi':   'http://www.w3.org/2001/XInclude',
+    'svg':  'http://www.w3.org/2000/svg',
     'xml':  'http://www.w3.org/XML/1998/namespace',
-    's':    'http://www.w3.org/2000/svg',
+    'xlink':'http://www.w3.org/1999/xlink',
 }
 
 
+def sdict(d):
+    "Create an ordered dict from the given dict, with sorted initial insertion"
+    # For python 3.6 and later, lxml obeys dictionary insertion order for
+    # attributes, for earlier it sorts initial attributes, so we need to
+    # use a regular dict for 3.6 and later
+    if sys.version_info.major == 3 and sys.version_info.minor >= 6:
+        return dict( (k, d[k]) for k in sorted(list(d.keys())) )
+    else:
+        if isinstance(d, OrderedDict):
+            return d
+        return OrderedDict( (k, d[k]) for k in sorted(list(d.keys())) )
+
 # ----------------------------------------------------------------------
 # Element operations
 
@@ -574,7 +623,7 @@ def isempty(e):
 
 def isblock(e):
     "Return True if e is a block level element"
-    return e.tag in [ 'artwork', 'dl', 'figure', 'ol', 'sourcecode', 't', 'ul', 'blockquote', ]
+    return e.tag in [ 'artset', 'artwork', 'dl', 'figure', 'ol', 'sourcecode', 't', 'table', 'ul', 'blockquote', ]
 
 def iscomment(e):
     "Return True if e is a comment"
@@ -588,4 +637,6 @@ def hastext(e):
 
 def is_htmlblock(h):
     return h.tag in set([ 'address', 'article', 'aside', 'blockquote', 'div', 'dl', 'figure',
-        'h1', 'h2', 'h3', 'h4', 'h5', 'h6', 'nav', 'ol', 'p', 'pre', 'section', 'table', 'ul', ])
+        'h1', 'h2', 'h3', 'h4', 'h5', 'h6', 'nav', 'ol', 'p', 'pre', 'script', 'section',
+        'table', 'ul', ])
+
