@@ -9,53 +9,74 @@ import textwrap
 import lxml
 import os
 import re
-import sys
 import xml2rfc.log
 import xml2rfc.util
 import xml2rfc.utils
 from optparse import Values
 
 try:
-    import debug
+    from xml2rfc import debug
     debug.debug = True
 except ImportError:
     pass
 
-from xml2rfc import strings
+from xml2rfc import strings, log
 from xml2rfc.util.date import extract_date, format_date, get_expiry_date
-from xml2rfc.util.name import short_author_ascii_name_parts, full_author_name_expansion
+from xml2rfc.util.name import short_author_ascii_name_parts, full_author_name_expansion, short_author_name_parts
 from xml2rfc.util.unicode import punctuation, unicode_replacements, unicode_content_tags, downcode
-from xml2rfc.utils import namespaces
+from xml2rfc.utils import namespaces, find_duplicate_ids
 
 default_options = Values(defaults={
         'accept_prepped': None,
         'add_xinclude': None,
         'basename': None,
+        'bom': False,
         'cache': None,
+        'clear_cache': False,
         'css': None,
+        'date': datetime.date.today(),
         'datestring': None,
         'debug': False,
         'doi_base_url': 'https://doi.org/',
         'dtd': None,
+        'expand': False,
         'external_css': False,
         'filename': None,
-        'id_base_url': 'https://www.ietf.org/archive/id/',
+        'first_page_author_org': True,
+        'html': False,
+        'id_base_url': 'https://tools.ietf.org/html/',
+        'id_reference_base_url': 'https://tools.ietf.org/html/',
+        'id_is_work_in_progress': True,
+        'image_svg': False,
         'indent': 2,
-        'legacy': False,
-        'legacy_date_format': True,
+        'info': False,
+        'legacy': True,
+        'legacy_date_format': False,
         'legacy_list_symbols': False,
         'list_symbols': ('*', '-', 'o', '+'),
+        'metadata_js_url': 'metadata.min.js',
+        'no_css': False,
         'no_dtd': None,
         'no_network': False,
+        'nroff': False,
         'omit_headers': None,
         'output_filename': None,
         'output_path': None,
+        'pagination': True,
+        'pdf': False,
+        'preptool': False,
         'quiet': False,
         'remove_pis': False,
-        'rfc_base_url': 'https://www.rfc-editor.org/info/',
+        'raw': False,
+        'rfc': None,
+        'rfc_base_url': 'https://www.rfc-editor.org/rfc/',
+        'rfc_reference_base_url': 'https://rfc-editor.org/rfc/',
         'strict': False,
+        'text': True,
         'utf8': False,
         'verbose': False,
+        'version': False,
+        'v2v3': False,
         'vocabulary': 'v2',
     })
 
@@ -376,6 +397,7 @@ class BaseRfcWriter:
         'trust200811',
         'noModificationTrust200811',
         'noDerivativesTrust200811',
+        'none',
     ]
 
     # -------------------------------------------------------------------------
@@ -684,11 +706,11 @@ class BaseRfcWriter:
         buf = []
         for i, author in enumerate(authors):
             organization = author.find('organization')
-            surname = author.attrib.get('surname', '')
+            initials, surname = short_author_name_parts(author)
             if i == len(authors) - 1 and len(authors) > 1:
                 buf.append('and ')
             if surname:
-                initials = self.get_initials(author)
+                initials = self.get_initials(author) or initials or ''
                 if i == len(authors) - 1 and len(authors) > 1:
                     # Last author is rendered in reverse
                     if len(initials) > 0:
@@ -782,8 +804,6 @@ class BaseRfcWriter:
     def _prepare_top_right(self):
         """ Returns a list of lines for the top right header """
         lines = []
-        # Render author?
-        stream = self.r.attrib.get('submissionType', self.defaults['submissionType'])
         # Keep track of previous organization and remove if redundant.
         last_org = None
         last_pos = None
@@ -798,7 +818,7 @@ class BaseRfcWriter:
                          get('surname', '') + role)
             organization = author.find('organization')
             org_name = ''
-            if stream != 'IAB':
+            if self.options.first_page_author_org:
                 if organization is not None:
                     abbrev = organization.attrib.get("abbrev", None)
                     if  abbrev != None and abbrev.strip() != '':
@@ -1221,12 +1241,13 @@ class BaseRfcWriter:
                 refs_newcounter = refs_counter + '.' + str(len(references)+1)
                 self._indexReferences(refs_newcounter, title="URIs", level=2, subCounter=len(references)+1)
 
-        for reference_list in references:
-            for ref in reference_list:
-                if len(ref):
-                    ref_counter += 1
-                    title = ref.find("front/title").text
-                    self._indexRef(ref_counter, title=title, anchor=ref.attrib["anchor"])
+
+        for ref in self.r.xpath('.//references//reference'):
+            if len(ref):
+                ref_counter += 1
+                title = ref.find("front/title")
+                if title != None:
+                    self._indexRef(ref_counter, title=title.text, anchor=ref.attrib["anchor"])
 
         # Appendix sections
         back = self.r.find('back')
@@ -1425,7 +1446,7 @@ class BaseRfcWriter:
             self.write_to_file(tmpfile)
 
         if not self.options.quiet and filename:
-            xml2rfc.log.write('Created file', filename)
+            xml2rfc.log.write(' Created file', filename)
 
     def write_erefs(self, refs_counter, refs_subsection):
         """ Only text versions do this so provide a default that does nothing
@@ -1443,7 +1464,7 @@ class BaseRfcWriter:
         if not self.indexmode:
             for item in self._index:
                 if item.autoAnchor.startswith("rfc.ref.") and not item.used:
-                    xml2rfc.log.warn("no <xref> in <rfc> targets <reference anchor='%s'>" % item.anchor)
+                    xml2rfc.log.warn("No <xref> in <rfc> targets <reference anchor='%s'>" % item.anchor)
 
     def needLines(self, count):
         """ Deal with the needLines PI """
@@ -1533,7 +1554,8 @@ class BaseRfcWriter:
         """ Writes the finished buffer to a file """
         raise NotImplementedError('write_to_file() needs to be overridden')
 
-
+# --------------------------------------------------------------------------------------------------
+# Sets of various kinds of tags in the v3 schema
         
 deprecated_element_tags = set([
     'c',
@@ -1548,69 +1570,145 @@ deprecated_element_tags = set([
     'vspace',
 ])
 
+v3_rng_file = os.path.join(os.path.dirname(os.path.dirname(__file__)), 'data', 'v3.rng')
+v3_schema = lxml.etree.ElementTree(file=v3_rng_file)
+
+def get_element_tags():
+    tags = set()
+    elements = v3_schema.xpath("/x:grammar/x:define/x:element", namespaces=namespaces)
+    for element in elements:
+        name = element.get('name')
+        if not name in tags:
+            tags.add(name)
+    return tags
+
+def get_meta_tags():
+    "Get tags that can have text content but don't occur in regular text"
+    tags = set()
+    for elem in ['author', 'address', 'front', 'postal', 'reference' ]:
+        refs = v3_schema.xpath("/x:grammar/x:define[@name='%s']//x:ref"%elem, namespaces=namespaces)
+        for r in refs:
+            name = r.get('name')
+            if not name in tags:
+                tags.add(name)
+    return tags
+
+def get_text_tags():
+    "Get tags that can have text content from the schema"
+    tags = set()
+    texts = v3_schema.xpath("/x:grammar/x:define/x:element//x:text", namespaces=namespaces)
+    for t in texts:
+        element = list(t.iterancestors('{*}element'))[0]
+        name = element.get('name')
+        if not name in tags:
+            tags.add(name)
+    return tags
+
+def get_text_or_block_tags():
+    "Get tags that can have either text or block contents from the schema"
+    tags = set()
+    texts = v3_schema.xpath("/x:grammar/x:define/x:element/x:choice/*/x:choice//x:text", namespaces=namespaces)
+    for t in texts:
+        element = list(t.iterancestors('{*}element'))[0]
+        name = element.get('name')
+        if not name in tags:
+            tags.add(name)
+    return tags
+
+def get_inline_tags():
+    "Get tags that can occur within text from the schema"
+    tags = set()
+    referenced = v3_schema.xpath("/x:grammar/x:define/x:element//x:ref", namespaces=namespaces)
+    for ref in referenced:
+        name = ref.get('name')
+        if not name in tags:
+            p = ref.getparent()
+            text = p.find('x:text', namespaces=namespaces)
+            if text != None:
+                tags.add(name)
+    return tags
+
+meta_tags   = get_meta_tags() - deprecated_element_tags
+text_tags   = get_text_tags() - deprecated_element_tags
+inline_tags = get_inline_tags() - deprecated_element_tags
+block_tags  = get_element_tags() - text_tags - deprecated_element_tags
+mixed_tags  = get_text_or_block_tags() - deprecated_element_tags
+
+# --------------------------------------------------------------------------------------------------
 
 class BaseV3Writer(object):
 
     def __init__(self, xmlrfc, quiet=None, options=default_options, date=datetime.date.today()):
+        global v3_rng_file, v3_schema
         self.xmlrfc = xmlrfc
-        self.tree = xmlrfc.tree
-        self.root = self.tree.getroot()
+        self.tree = xmlrfc.tree if xmlrfc else None
+        self.root = self.tree.getroot() if xmlrfc else None
         self.options = options
         self.date = date
+        self.rng_file = v3_rng_file
+        self.v3_rng = None
+        self.schema = v3_schema
         self.index_items = []
-        self.v3_rng_file = os.path.join(os.path.dirname(os.path.dirname(__file__)), 'data', 'v3.rng')
-        self.schema = lxml.etree.ElementTree(file=self.v3_rng_file)
-        self.text_tags   = self.get_text_tags() - deprecated_element_tags
-        self.inline_tags = self.get_inline_tags() - deprecated_element_tags
+        self.meta_tags = set(meta_tags)
+        self.text_tags = set(text_tags)
+        self.inline_tags = set(inline_tags)
+        self.mixed_tags = set(mixed_tags)
         self.refname_mapping = dict( (e.get('anchor'), e.get('anchor')) for e in self.root.xpath('.//reference') )
         self.refname_mapping.update(dict( (e.get('target'), e.get('to')) for e in self.root.xpath('.//displayreference') ))
+        self.refname_mapping.update(dict( (e.get('anchor'), e.get('anchor')) for e in self.root.xpath('.//referencegroup') ))
         #
         self.errors = []
-        if options.debug:
-            found_handlers = []
-            missing_handlers = []
-            for tag in self.element_tags:
-                func_name = "render_%s" % (tag.lower(),)
-                if getattr(self, func_name, False):
-                    found_handlers.append(func_name)
-                else:
-                    missing_handlers.append(func_name)
-            debug.pprint('found_handlers')
-            debug.show('len(found_handlers)')
-            debug.pprint('missing_handlers')
-            debug.show('len(missing_handlers)')
+#         if options.debug:
+#             found_handlers = []
+#             missing_handlers = []
+#             for tag in (self.get_element_tags() - deprecated_element_tags):
+#                 func_name = "render_%s" % (tag.lower(),)
+#                 if getattr(self, func_name, False):
+#                     found_handlers.append(func_name)
+#                 else:
+#                     missing_handlers.append(func_name)
+#             debug.pprint('found_handlers')
+#             debug.show('len(found_handlers)')
+#             debug.pprint('missing_handlers')
+#             debug.show('len(missing_handlers)')
+
+    def log(self, msg):
+        xml2rfc.log.write(msg)
 
     def msg(self, e, label, text):
-        lnum = getattr(e, 'sourceline', None)
-        file = getattr(e, 'base', None)
-        if lnum:
-            msg = "%s(%s): %s %s" % (file or self.xmlrfc.source, lnum, label, text, )
+        if e != None:
+            lnum = getattr(e, 'sourceline', None)
+            file = getattr(e, 'base', None)
+            if lnum:
+                msg = "%s(%s): %s %s" % (file or self.xmlrfc.source, lnum, label, text, )
+            else:
+                msg = "(No source line available): %s %s" % (label, text, )
         else:
-            msg = "(No source line available): %s %s" % (label, text, )
-        xml2rfc.log.write(msg)
+            msg = "%s %s" % (label, text)
         return msg
 
     def get_relevant_pis(self, e):
         pis = []
-        # directly inside element
-        for c in e.getchildren():
-            if c.tag == lxml.etree.PI and c.target == xml2rfc.V3_PI_TARGET:
-                pis.append(c)
-        # siblings before element
-        for s in e.itersiblings(preceding=True):
-            if s.tag == lxml.etree.PI and s.target == xml2rfc.V3_PI_TARGET:
-                pis.append(s)
-        # ancestor's earlier siblings
-        for a in e.iterancestors():
-            for s in a.itersiblings(preceding=True):
+        if e != None:
+            # directly inside element
+            for c in e.getchildren():
+                if c.tag == lxml.etree.PI and c.target == xml2rfc.V3_PI_TARGET:
+                    pis.append(c)
+            # siblings before element
+            for s in e.itersiblings(preceding=True):
                 if s.tag == lxml.etree.PI and s.target == xml2rfc.V3_PI_TARGET:
                     pis.append(s)
-        # before root elements
-        p = self.root.getprevious()
-        while p != None:
-            if p.tag == lxml.etree.PI and p.target == xml2rfc.V3_PI_TARGET:
-                pis.append(p)
-            p = p.getprevious()
+            # ancestor's earlier siblings
+            for a in e.iterancestors():
+                for s in a.itersiblings(preceding=True):
+                    if s.tag == lxml.etree.PI and s.target == xml2rfc.V3_PI_TARGET:
+                        pis.append(s)
+            # before root elements
+            p = self.root.getprevious()
+            while p != None:
+                if p.tag == lxml.etree.PI and p.target == xml2rfc.V3_PI_TARGET:
+                    pis.append(p)
+                p = p.getprevious()
         return pis
 
     def silenced(self, e, text):
@@ -1631,80 +1729,78 @@ class BaseV3Writer(object):
         return False
 
     def note(self, e, text, label='Note:'):
-        if not self.options.quiet:
+        if self.options.verbose:
             if not self.silenced(e, text):
-                self.msg(e, label, text)
+                self.log(self.msg(e, label, text))
 
     def warn(self, e, text):
         if not self.silenced(e, text):
-            self.msg(e, 'Warning:', text)
+            self.log(self.msg(e, 'Warning:', text))
 
     def err(self, e, text, trace=False):
         msg = self.msg(e, 'Error:', text)
+        self.errors.append(msg)
         if trace or self.options.debug:
             raise RuntimeError(msg)
-        self.errors.append(msg)
+        else:
+            self.log(msg)
 
     def die(self, e, text, trace=False):
-        self.err(e, text, trace)
-        xml2rfc.log.write("Cannot continue, quitting now")
-        sys.exit(1)
+        msg = self.msg(e, 'Error:', text)
+        self.errors.append(msg)
+        raise RfcWriterError(msg)
+
+    # methods operating on the xml tree
+
+    def get_element_from_id(self, id):
+        for a in 'anchor', 'pn', 'slugifiedName':
+            elem = self.root.find('.//*[@%s="%s"]'%(a, id, ))
+            if elem != None:
+                break
+        return elem
+
+    def get_element_page(self, e):
+        page = getattr(e, 'page', None)
+        if not page:
+            for a in e.iterancestors():
+                page = getattr(a, 'page', None)
+                if page != None:
+                    break
+        return page
 
 #     def get_valid_child_tags(self, tag):
 #         refs = self.schema.xpath("/x:grammar/x:define/x:element[@name='%s']//x:ref" % tag, namespaces=utils.namespaces)
 #         names = set([ r.get('name') for r in refs ])
 #         return names
 
-    def get_text_tags(self):
-        "Get tags that can have text content from the schema"
-        text_tags = set()
-        texts = self.schema.xpath("/x:grammar/x:define/x:element//x:text", namespaces=namespaces)
-        for t in texts:
-            element = list(t.iterancestors('{*}element'))[0]
-            name = element.get('name')
-            if not name in text_tags:
-                text_tags.add(name)
-        return text_tags
-
-    def get_inline_tags(self):
-        "Get tags that can occur within text from the schema"
-        inline_tags = set()
-        referenced = self.schema.xpath("/x:grammar/x:define/x:element//x:ref", namespaces=namespaces)
-        for ref in referenced:
-            name = ref.get('name')
-            if not name in inline_tags:
-                p = ref.getparent()
-                text = p.find('x:text', namespaces=namespaces)
-                if text != None:
-                    inline_tags.add(name)
-        return inline_tags
-
     def write(self, filename):
         raise NotImplementedError()
 
     # methods for use in setting up page headers and footers
 
-    def footer_series(self):
-        if   self.root.find('.//seriesInfo[@name="Internet-Draft"]') != None:
-            return 'Internet-Draft'
-        else:
-            info = self.root.find('.//seriesInfo[@name="RFC"]')
-            if info != None:
-                return 'RFC %s' % info.get('value')
-        return '(unknown)'
+    def page_top_left(self):
+        if self.root.get('ipr') == 'none':
+            return ''
+        number = self.root.get('number')
+        if number != None:
+            return 'RFC %s' % number
+        info = self.root.find('./front/seriesInfo[@name="RFC"]')
+        if info != None:
+            return 'RFC %s' % info.get('value')
+        return 'Internet-Draft'
 
-    def footer_title(self):
+    def page_top_center(self):
         title = self.root.find('./front/title')
-        text = title.get('abbrev') or ''.join(title.itertext())
+        text = title.get('abbrev') or ' '.join(title.itertext())
         return text
 
-    def footer_date(self):
+    def page_top_right(self):
         date = self.root.find('./front/date')
         parts = extract_date(date, self.options.date)
         text = format_date(parts[0], parts[1], None, legacy=True)
         return text
 
-    def footer_authors(self):
+    def page_bottom_left(self):
         authors = self.root.findall('front/author')
         authors = [ a for a in authors if a.get('role') != 'contributor' ]
         surnames = [ n for n in [ short_author_ascii_name_parts(a)[1] for a in authors ] if n ]
@@ -1716,9 +1812,9 @@ class BaseV3Writer(object):
             text = '%s, et al.' % surnames[0]
         return text
 
-    def footer_center(self):
+    def page_bottom_center(self):
         # Either expiry date or category
-        if self.options.rfc:
+        if self.options.rfc or self.root.get('ipr') == 'none':
             cat = self.root.get('category')
             text = strings.category_name[cat]
         else:
@@ -1772,12 +1868,12 @@ class BaseV3Writer(object):
                         e.tail = '\n'+' '*(i-ind)
                 return
             #
-            if e.tag not in self.text_tags:
+            if e.tag not in text_tags:
                 if len(e) and (e.text is None or e.text.strip()==''):
                     e.text = '\n'+' '*(i+ind)
             elif e.tag in ['blockquote', 'li', 'dd', 'td', 'th' ]: # mixed content
                 pass
-                if len(e) and e[0] not in self.inline_tags and (e.text is None or e.text.strip()==''):
+                if len(e) and e[0] not in inline_tags and (e.text is None or e.text.strip()==''):
                     e.text = '\n'+' '*(i+ind)
             elif e.tag in ['artwork', 'sourcecode', ]:
                 pass
@@ -1797,7 +1893,7 @@ class BaseV3Writer(object):
             for c in e:
                 indent(c, i+ind)
             #
-            if e.tag not in self.inline_tags:# and e.tag not in ['artwork', 'sourcecode', ]:
+            if e.tag not in inline_tags:# and e.tag not in ['artwork', 'sourcecode', ]:
                 if e.tail is None or e.tail.strip()=='':
                     if e.getnext() != None:
                         e.tail = '\n'+' '*i
@@ -1813,3 +1909,59 @@ class BaseV3Writer(object):
                         e.tail = '\n'+' '*(i-ind)
         indent(e, 0)
         e.tail = None
+
+    def validate(self, when='', warn=False):
+        # Note: Our schema doesn't permit xi:include elements, so the document
+        # must have had XInclude processing done before calling validate()
+
+        # The lxml Relax NG validator checks that xsd:ID values are unique,
+        # but unfortunately the error messages are completely unhelpful (lxml
+        # 4.1.1, libxml 2.9.1): "Element li has extra content: t" when 't' has
+        # a duplicate xsd:ID attribute.  So we check all attributes with
+        # content specified as xsd:ID first, and give better messages:
+
+        # Get the attributes we need to check
+        if when and not when.startswith(' '):
+            when = ' '+when
+        dups = find_duplicate_ids(self.schema, self.tree)
+        for attr, id, e in dups:
+            self.warn(e, 'Duplicate xsd:ID attribute %s="%s" found.  This will cause validation failure.' % (attr, id, ))
+
+        try:
+            if not self.v3_rng:
+                self.v3_rng = lxml.etree.RelaxNG(file=self.rng_file)
+                
+            self.v3_rng.assertValid(self.tree)
+            return True
+        except Exception as e:
+            lxmlver = lxml.etree.LXML_VERSION[:3]
+            if lxmlver < (3, 8, 0):
+                self.warn(None, "The available version of the lxml library (%s) does not provide xpath "
+                          "information as part of validation errors.  Upgrade to version 3.8.0 or "
+                          "higher for better error messages." % ('.'.join(str(v) for v in lxmlver), ))
+            # These warnings are occasionally incorrect -- disable this
+            # output for now:
+            if hasattr(e, 'error_log'):
+                for error in e.error_log:
+                    path = getattr(error, 'path', '')
+                    msg = "%s(%s): %s: %s, at %s" % (self.xmlrfc.source, error.line, error.level_name.title(), error.message, path)
+                    self.log(msg)
+            else:
+                log.warn('\nInvalid document: %s' % (e,))
+            if warn:
+                self.warn(self.root, 'Invalid document%s.' % (when, ))
+                return False
+            else:
+                self.die(self.root, 'Invalid document%s.' % (when, ))
+
+    def remove(self, p, e):
+        # Element.remove(child) removes both the child and its tail, so in
+        # order not to loose text when removing comments or other elements,
+        # we need to handle that case:
+        if e.tail:
+            s = e.getprevious()
+            if s != None:
+                s.tail = (s.tail or '') + e.tail
+            else:
+                p.text = (p.text or '') + e.tail
+        p.remove(e)
