@@ -8,18 +8,20 @@ import lxml.etree
 import datetime
 import traceback as tb
 
+from collections import OrderedDict
 from io import open
 from lxml import etree
 from lxml.etree import Element, Comment, CDATA
 
+import xml2rfc
 from xml2rfc import log
 from xml2rfc.util.unicode import unicode_content_tags, isascii
-from xml2rfc.utils import find_duplicate_ids, hastext, isempty
+from xml2rfc.utils import hastext, isempty, sdict
 from xml2rfc.writers.base import default_options, BaseV3Writer
 
 
 try:
-    import debug
+    from xml2rfc import debug
     assert debug
 except ImportError:
     pass
@@ -61,16 +63,7 @@ class V2v3XmlWriter(BaseV3Writer):
         self.schema = etree.ElementTree(file=self.v3_rng_file)
 
     def validate(self):
-        v3_rng_file = os.path.join(os.path.dirname(os.path.dirname(__file__)), 'data', 'v3.rng')
-        v3_rng = lxml.etree.RelaxNG(file=v3_rng_file)
-        try:
-            v3_rng.assertValid(self.root)
-            log.note("The document validates according to the RFC7991 schema")
-        except Exception as e:
-            dups = find_duplicate_ids(self.schema, self.tree)
-            for attr, id, e in dups:
-                self.warn(e, 'Duplicate xsd:ID attribute %s="%s" found.  This will cause validation failure.' % (attr, id, ))
-            log.warn('\nInvalid document: %s' % (e,))
+        return super(V2v3XmlWriter, self).validate(when='when running v2v3 conversion', warn=True)
 
     def add_xinclude(self):
         for e in self.root.xpath('.//back//reference'):
@@ -82,7 +75,8 @@ class V2v3XmlWriter(BaseV3Writer):
                         xi = self.element('{http://www.w3.org/2001/XInclude}include',
                                     nsmap=self.xmlrfc.nsmap,
                                     line=e.sourceline,
-                                    href="https://xml2rfc.tools.ietf.org/public/rfc/bibxml/reference.RFC.%s.xml"%num)
+                                    href="https://xml2rfc.tools.ietf.org/public/rfc/bibxml/reference.RFC.%04d.xml"%int(num))
+                        xi.tail = e.tail
                         i = p.index(e)
                         p.remove(e)
                         p.insert(i, xi)
@@ -92,16 +86,23 @@ class V2v3XmlWriter(BaseV3Writer):
                     name = si.get('value', '')
                     if name:
                         tag = name[len('draft-'):] if name.startswith('draft-') else name
-                        xi = self.element('{http://www.w3.org/2001/XInclude}include',
-                                    nsmap=self.xmlrfc.nsmap,
-                                    line=e.sourceline,
-                                    href="https://xml2rfc.tools.ietf.org/publicrfc/bibxml3/reference.I-D.%s.xml"%tag)
+                        if re.search(r'-\d\d$', tag):
+                            xi = self.element('{http://www.w3.org/2001/XInclude}include',
+                                        nsmap=self.xmlrfc.nsmap,
+                                        line=e.sourceline,
+                                        href="https://xml2rfc.tools.ietf.org/public/rfc/bibxml3/reference.I-D.draft-%s.xml"%tag)
+                        else:
+                            xi = self.element('{http://www.w3.org/2001/XInclude}include',
+                                        nsmap=self.xmlrfc.nsmap,
+                                        line=e.sourceline,
+                                        href="https://xml2rfc.tools.ietf.org/public/rfc/bibxml3/reference.I-D.%s.xml"%tag)
+                        xi.tail = e.tail
                         i = p.index(e)
                         p.remove(e)
                         p.insert(i, xi)
                     else:
                         self.warn(e, 'Invalid value in %s' % lxml.etree.tostring(e))
-        lxml.etree.cleanup_namespaces(self.root, top_nsmap=self.xmlrfc.nsmap)
+#        lxml.etree.cleanup_namespaces(self.root, top_nsmap=self.xmlrfc.nsmap)
 
 
     def post_process_lines(self, lines):
@@ -112,25 +113,34 @@ class V2v3XmlWriter(BaseV3Writer):
         """ Public method to write the XML document to a file """
 
         self.convert2to3()
-        self.validate()
         if self.options.add_xinclude:
             self.add_xinclude()
 
         # Use lxml's built-in serialization
         file = open(filename, 'w', encoding='utf-8')
         text = lxml.etree.tostring(self.root.getroottree(), 
-                                       xml_declaration=True, 
-                                       encoding='utf-8',
+                                       encoding='unicode',
+                                       doctype='<!DOCTYPE rfc SYSTEM "rfc2629-xhtml.ent">',
                                        pretty_print=True)
-        file.write(text.decode('utf-8'))
+
+        # Use entities for some selected unicode code points, for later
+        # editing readability and convenience
+        text = text.replace(u'\u00A0', u'&nbsp;')
+        text = text.replace(u'\u200B', u'&zwsp;')
+        text = text.replace(u'\u2011', u'&nbhy;')
+        text = text.replace(u'\u2028', u'&br;')
+        text = text.replace(u'\u2060', u'&wj;')
+
+        file.write(u"<?xml version='1.0' encoding='utf-8'?>\n")
+        file.write(text)
 
         if not self.options.quiet:
-            log.write('Created file', filename)
+            self.log(' Created file %s' % filename)
 
     # --- Element Operations -------------------------------------------
 
     def element(self, tag, line=None, **kwargs):
-        e = Element(tag, **kwargs)
+        e = Element(tag, **sdict(kwargs))
         if line:
             e.sourceline = line
         elif self.options.debug:
@@ -172,7 +182,8 @@ class V2v3XmlWriter(BaseV3Writer):
                     b.text = a.text
                 if a.tail != None:
                     b.tail = a.tail
-                b.sourceline = a.sourceline
+                if a.sourceline:
+                    b.sourceline = a.sourceline
                 copyattr(a, b)
                 for child in a.iterchildren():
                     b.append(child)         # moves child from a to b
@@ -184,7 +195,7 @@ class V2v3XmlWriter(BaseV3Writer):
                     else:
                         c.tail = a.tail
                 p.remove(a)
-        if b != None:
+        if b != None and a.sourceline:
             b.sourceline = a.sourceline
         return b
 
@@ -237,7 +248,8 @@ class V2v3XmlWriter(BaseV3Writer):
         if t is None:
             t = self.element('t', line=e.sourceline)
         t.text = e.text
-        t.sourceline = e.sourceline
+        if e.sourceline:
+            t.sourceline = e.sourceline
         e.text = None
         for c in e.iterchildren():
             t.append(c)
@@ -249,6 +261,7 @@ class V2v3XmlWriter(BaseV3Writer):
     def convert2to3(self):
         if self.root.get('version') in ['3', ]:
             return self.tree
+        log.note(' Converting v2 to v3: %s' % self.xmlrfc.source)
 
         selectors = [
             # we need to process list before block elements that might get
@@ -335,6 +348,19 @@ class V2v3XmlWriter(BaseV3Writer):
 
         self.root.set('version', '3')
 
+        # Add a comment about the converter version
+        conversion_version = Comment(' xml2rfc v2v3 conversion %s ' % xml2rfc.__version__)
+        conversion_version.tail = '\n  '
+        self.root.insert(0, conversion_version)
+
+        # This is a workaround for not being able to do anything about
+        # namespaces other than when creating an element.  It lets us retain
+        # a namespace declaration for xi: in the root element.
+#         dummy = self.element('{http://www.w3.org/2001/XInclude}include', nsmap=self.xmlrfc.nsmap)
+#         self.root.insert(0, dummy)
+#         lxml.etree.cleanup_namespaces(self.root, top_nsmap=self.xmlrfc.nsmap, keep_ns_prefixes='xi')
+#         self.root.remove(dummy)
+        log.note(' Completed v2 to v3 conversion')
         return self.tree
 
     # ----------------------------------------------------------------------
@@ -361,6 +387,7 @@ class V2v3XmlWriter(BaseV3Writer):
                     v = 'true'
                 rfc_element.set(a, v)
                 self.replace(e, None, 'Moved %s PI to <rfc %s="%s"' % (k, a, v))
+                break
         else:
             self.replace(e, None, 'Removed %s PI"' % (k,))
 
@@ -410,7 +437,7 @@ class V2v3XmlWriter(BaseV3Writer):
     # 
     #    Deprecated.
     def element_artwork(self, e, p):
-        if e.text and e.text.strip():
+        if e.text and e.text.strip() and not ']]>' in e.text:
             e.text = CDATA(e.text)          # prevent text from being mucked up by other processors
         stripattr(e, ['height', '{http://www.w3.org/XML/1998/namespace}space', 'width', ])
         if e.text and re.search(r'^\s*<CODE BEGINS>', e.text):
@@ -433,6 +460,22 @@ class V2v3XmlWriter(BaseV3Writer):
                 self.warn(e, "Found non-whitespace content after <CODE ENDS>")
                 e.tail = tail
             stripattr(e, ['align', 'alt', 'height', 'suppress-title', 'width', ])
+        src = e.get('src')
+        if e.text and e.text.strip() and src:
+            # We have both text content and a src attribute -- convert to
+            # <artset> with 2 <artwork)
+            ext = os.path.splitext(src)[1][1:]
+            artset = self.element('artset', line=e.sourceline)
+            extart = self.copy(e, 'artwork')
+            extart.text = None
+            extart.tail = None
+            extart.set('type', ext)
+            artset.append(extart)
+            stripattr(e, ['src'])
+            e.set('type', 'ascii-art')
+            artset.append(e)
+            p.append(artset)
+
 
     def element_back(self, e, p):
         # XXXX The RFCs don't say anything about the text rendering of
@@ -489,7 +532,7 @@ class V2v3XmlWriter(BaseV3Writer):
     # 
     def element_figure(self, e, p):
         comments = []
-        embedded = e.find('./artwork')
+        embedded = e.find('.//artwork')
         if embedded == None:
             embedded = e.find('./sourcecode')
         if embedded != None:
@@ -510,8 +553,11 @@ class V2v3XmlWriter(BaseV3Writer):
             and e.find('name')==None and e.find('preamble')==None and e.find('postamble')==None ):
             pos = p.index(e)
             embedded.tail = e.tail
+            children = e.getchildren()
             p.remove(e)
-            p.insert(pos, embedded)
+            for c in children:
+                p.insert(pos, c)
+                pos += 1
         else:
             stripattr(e, ['align', 'height', 'src', 'suppress-title', 'width', ])
             if self.options.strict:
@@ -588,6 +634,8 @@ class V2v3XmlWriter(BaseV3Writer):
 #         }
         def equal(e1, e2):
             return (e1.get('name'), e1.get('value')) ==  (e2.get('name'), e2.get('value'))
+        if e.get('ipr') == 'none':
+            return
         front = e.find('./front')
         title = e.find('./front/title')
         i = front.index(title) + 1 if title!=None else 0
@@ -609,10 +657,6 @@ class V2v3XmlWriter(BaseV3Writer):
                 e.insert(0, self.element('link', rel='prev', href="https://datatracker.ietf.org/doc/%s"%(e.get('docName'), ), line=e.sourceline))
         elif 'docName' in e.attrib:
             value=e.get('docName')
-            if '.' in value:
-                log.warn("The 'docName' attribute of the <rfc/> element should not contain any filename extension: docName=\"draft-foo-bar-02\".")
-            if not re.search('-\d\d$', value):
-                log.warn("The 'docName' attribute of the <rfc/> element should have a revision number as the last component when submitted: docName=\"draft-foo-bar-02\".")
             new = self.element('seriesInfo', name="Internet-Draft", value=value, line=e.sourceline)
             if not 'Internet-Draft' in [ s.get('name') for s in series ]:
                 front.insert(i, new)
@@ -653,8 +697,6 @@ class V2v3XmlWriter(BaseV3Writer):
     #       <xref> because it makes no sense semantically.
     #
     def element_xref(self, e, p):
-        if 'format' in e.attrib and e.get('format') == 'none':
-            stripattr(e, ['format'])
         stripattr(e, ['pageno'])
 
     # 2.66.2.  "pageno" Attribute
@@ -731,7 +773,7 @@ class V2v3XmlWriter(BaseV3Writer):
         # convert to dl, ul, or ol
         nstyle = None
         style = e.get('style', '').strip()
-        attribs = {}
+        attribs = OrderedDict()
         comments = []
         if not style:
             # otherwise look for the nearest list parent with a style and use it
@@ -1034,6 +1076,7 @@ class V2v3XmlWriter(BaseV3Writer):
                 l.remove(t)
         else:
             self.replace(e, None, "<vspace/> deprecated and removed")
+            self.warn(e, "Deprecated <vspace/> element removed, but no good conversion found  The output will most likely need fixup.")
 
     # 3.10.1.  "blankLines" Attribute
     # 
